﻿using System;
using System.Linq;
using System.Threading;

namespace Barrier
{
    class Program
    {

        const int threads = 10;

        /// <summary>
        /// Basic function that prints letter by letter in multiple threads. No synchornization
        /// </summary>
        /// <param name="inputstring"></param>
        static void printString(string inputstring)
        {
            
            ThreadStart thread = () =>
            {
                char[] inputArray = inputstring.ToArray();
                for (int i = 0; i < inputArray.Length; i++)
                {
                    Console.Write(inputArray[i]);
                }
            };
            Thread[] threadsArray = new Thread[threads];
            for (int i = 0; i < Program.threads; ++i)
            {
                threadsArray[i] = new Thread(thread);
                threadsArray[i].Start();
            }
            //Here we wait until all threads running in this method are finished
            for (int i = 0; i < threads; ++i)
            {
                threadsArray[i].Join();
            }
        }

        /// <summary>
        /// Checking out and synchronizing with Wait and PulseAll
        /// </summary>
        /// <param name="inputstring"></param>
        static void printStringWaitPulse(string inputstring)
        {
            object o = new object();
            int callCounter = 0;
            ThreadStart thread = () =>
            {
                char[] inputArray = inputstring.ToArray();
                for (int i = 0; i < inputArray.Length; i++)
                {
                    lock (o)
                    {
                        Console.Write(inputArray[i]);
                        callCounter++;
                        if (callCounter < threads)
                        {   //if not all threads finished we wait 
                            //and simultanously free the lock
                            Monitor.Wait(o);
                        }
                        else
                        {   //if we are the last thread we pulse others
                            Monitor.PulseAll(o); 
                            callCounter = 0;
                        }
                    }
                    //Monitor.Pulse(o);
                }
            };
            Thread[] threadsArray = new Thread[threads];
            for (int i = 0; i < threads; ++i)
            {
                threadsArray[i] = new Thread(thread);
                threadsArray[i].Start();
            }
            //Here we wait until all threads running in this method are finished
            for (int i = 0; i < threads; ++i)
            {
                threadsArray[i].Join();
            }
        }

        static CountdownEvent ce = new CountdownEvent(threads);

        //static System.Threading.Barrier bariera = new System.Threading.Barrier(threads);
        //in the second constructor parameter, you can specify the action to be performed 
        //when the barrier is achieved, eg displaying the counter.
        static System.Threading.Barrier barrier = new System.Threading.Barrier(threads, (b) => 
                    { Console.WriteLine(" Barrier in a phase: {0}", b.CurrentPhaseNumber); });
        /// <summary>
        /// Function demonstrating the operation of the barrier and CountdownEvent.
        /// the barrier is used to synchronize all threads after a certain stage is completed
        /// CountdownEvent waits for all threads to finish, so we move on. 
        /// </summary>
        /// <param name="inputstring"></param>
        static void printStringBarrier(string inputstring)
        {
            ThreadStart thread = () =>
            {
                for (int i = 0; i < inputstring.Length; i++)
                {
                    Console.Write(inputstring[i]);
                    barrier.SignalAndWait();
                }
                ce.Signal(); //when the task is done we report it to decrease the counter
            };
            Thread[] threadsArray = new Thread[threads];
            for (int i = 0; i < threads; ++i)
            {
                threadsArray[i] = new Thread(thread);
                threadsArray[i].Start();
            }
            //Here we wait until the counter reaches 0 (all threads calling ce.Signal ())
            ce.Wait();
            Console.WriteLine("Done. InitialCount={0}, CurrentCount={1}, IsSet={2}",
            ce.InitialCount, ce.CurrentCount, ce.IsSet);
        }

        static void Main(string[] args)
        {
            printString("All work and no play makes Jack a dull boy");
            Console.WriteLine();
            Console.WriteLine("----------"); //when we don't wait for the threads to end, this is even more of a mess
            printStringWaitPulse("All work and no play makes Jack a dull boy");
            Console.WriteLine();
            Console.WriteLine("----------");
            printStringBarrier("All work and no play makes Jack a dull boy");
            Console.WriteLine();

        }
    }
}
